"""
this is where calculations are performed on the 17Lands public data sets and
aggregate calculations are returned.

Aggregate dataframes containing raw counts are cached in the local file system
for performance.
"""

import datetime
import functools
import hashlib
import re
from typing import Callable, TypeVar

import polars as pl

from spells.external import data_file_path
import spells.cache
import spells.filter
import spells.manifest
from spells.columns import ColumnDefinition, ColumnSpec
from spells.enums import View, ColName, ColType


DF = TypeVar("DF", pl.LazyFrame, pl.DataFrame)


def _cache_key(args) -> str:
    """
    cache arguments by __str__ (based on the current value of a mutable, so be careful)
    """
    return hashlib.md5(str(args).encode("utf-8")).hexdigest()


@functools.lru_cache(maxsize=None)
def _get_names(set_code: str) -> tuple[str, ...]:
    card_fp = data_file_path(set_code, View.CARD)
    card_view = pl.read_parquet(card_fp)
    card_names_set = frozenset(card_view.get_column("name").to_list())

    draft_fp = data_file_path(set_code, View.DRAFT)
    draft_view = pl.scan_parquet(draft_fp)
    cols = draft_view.collect_schema().names()

    prefix = "pack_card_"
    names = tuple(col[len(prefix) :] for col in cols if col.startswith(prefix))
    draft_names_set = frozenset(names)

    assert (
        draft_names_set == card_names_set
    ), "names mismatch between card and draft file"
    return names


def _hydrate_col_defs(set_code: str, col_spec_map: dict[str, ColumnSpec]):
    def get_views(spec: ColumnSpec) -> list[View]:
        if spec.name == ColName.NAME or spec.col_type == ColType.AGG:
            return []
        if spec.col_type == ColType.CARD_ATTR:
            return [View.CARD]
        if spec.views is not None:
            return spec.views
        assert (
            spec.dependencies is not None
        ), f"Col {spec.name} should have dependencies"

        views = []
        for dep in spec.dependencies:
            views.extend(get_views(col_spec_map[dep]))

        return list(set(views))

    names = _get_names(set_code)
    assert len(names) > 0, "there should be names"
    hydrated = {}
    for key, spec in col_spec_map.items():
        if spec.col_type == ColType.NAME_SUM and spec.exprMap is not None:
            unnamed_exprs = map(spec.exprMap, names)
            expr = tuple(
                map(
                    lambda ex, name: ex.alias(f"{spec.name}_{name}"),
                    unnamed_exprs,
                    names,
                )
            )
        elif spec.expr is not None:
            expr = spec.expr.alias(spec.name)

        else:
            if spec.col_type == ColType.NAME_SUM:
                expr = tuple(map(lambda name: pl.col(f"{spec.name}_{name}"), names))
            else:
                expr = pl.col(spec.name)

        try:
            sig_expr = expr if isinstance(expr, pl.Expr) else expr[0]
            expr_sig = sig_expr.meta.serialize(
                format="json"
            )  # not compatible with renaming
        except pl.exceptions.ComputeError:
            if spec.version is not None:
                expr_sig = spec.name + spec.version
            else:
                expr_sig = str(datetime.datetime.now)

        dependencies = tuple(spec.dependencies or ())
        views = get_views(spec)
        signature = str(
            (
                spec.name,
                spec.col_type.value,
                expr_sig,
                tuple(view.value for view in views),
                dependencies,
            )
        )

        cdef = ColumnDefinition(
            name=spec.name,
            col_type=spec.col_type,
            views=tuple(views),
            expr=expr,
            dependencies=dependencies,
            signature=signature,
        )
        hydrated[key] = cdef
    return hydrated


def _view_select(
    df: DF,
    view_cols: frozenset[str],
    col_def_map: dict[str, ColumnDefinition],
    is_agg_view: bool,
) -> DF:
    base_cols = frozenset()
    cdefs = [col_def_map[c] for c in view_cols]
    select = []
    for cdef in cdefs:
        if is_agg_view:
            if cdef.col_type == ColType.AGG:
                base_cols = base_cols.union(cdef.dependencies)
                select.append(cdef.expr)
            else:
                base_cols = base_cols.union(frozenset({cdef.name}))
                select.append(cdef.name)
        else:
            if cdef.dependencies:
                base_cols = base_cols.union(cdef.dependencies)
            else:
                base_cols = base_cols.union(frozenset({cdef.name}))
            if isinstance(cdef.expr, tuple):
                select.extend(cdef.expr)
            else:
                select.append(cdef.expr)

    if base_cols != view_cols:
        df = _view_select(df, base_cols, col_def_map, is_agg_view)

    return df.select(select)


def _fetch_or_cache(
    calc_fn: Callable,
    set_code: str,
    cache_args,
    read_cache: bool = True,
    write_cache: bool = True,
):
    key = _cache_key(cache_args)

    if read_cache:
        if spells.cache.cache_exists(set_code, key):
            return spells.cache.read_cache(set_code, key)

    df = calc_fn()

    if write_cache:
        spells.cache.write_cache(set_code, key, df)

    return df


def _base_agg_df(
    set_code: str,
    m: spells.manifest.Manifest,
    use_streaming: bool = False,
) -> pl.DataFrame:
    join_dfs = []
    group_by = m.base_view_group_by

    is_name_gb = ColName.NAME in group_by
    nonname_gb = tuple(gb for gb in group_by if gb != ColName.NAME)

    for view, cols_for_view in m.view_cols.items():
        if view == View.CARD:
            continue
        df_path = data_file_path(set_code, view)
        base_view_df = pl.scan_parquet(df_path)
        base_df_prefilter = _view_select(
            base_view_df, cols_for_view, m.col_def_map, is_agg_view=False
        )

        if m.filter is not None:
            base_df = base_df_prefilter.filter(m.filter.expr)
        else:
            base_df = base_df_prefilter

        sum_cols = tuple(
            c
            for c in cols_for_view
            if m.col_def_map[c].col_type in (ColType.PICK_SUM, ColType.GAME_SUM)
        )
        if sum_cols:
            # manifest will verify that GAME_SUM manifests do not use NAME grouping
            name_col_tuple = (
                (pl.col(ColName.PICK).alias(ColName.NAME),) if is_name_gb else ()
            )

            sum_col_df = base_df.select(nonname_gb + name_col_tuple + sum_cols)
            join_dfs.append(
                sum_col_df.group_by(group_by).sum().collect(streaming=use_streaming)
            )

        name_sum_cols = tuple(
            c for c in cols_for_view if m.col_def_map[c].col_type == ColType.NAME_SUM
        )
        for col in name_sum_cols:
            cdef = m.col_def_map[col]
            pattern = f"^{cdef.name}_"
            name_map = functools.partial(
                lambda patt, name: re.split(patt, name)[1], pattern
            )

            expr = pl.col(f"^{cdef.name}_.*$").name.map(name_map)
            pre_agg_df = base_df.select((expr,) + nonname_gb)

            if nonname_gb:
                agg_df = pre_agg_df.group_by(nonname_gb).sum()
            else:
                agg_df = pre_agg_df.sum()

            index = nonname_gb if nonname_gb else None
            unpivoted = agg_df.unpivot(
                index=index,
                value_name=m.col_def_map[col].name,
                variable_name=ColName.NAME,
            )

            if not is_name_gb:
                df = (
                    unpivoted.drop("name")
                    .group_by(nonname_gb)
                    .sum()
                    .collect(streaming=use_streaming)
                )
            else:
                df = unpivoted.collect(streaming=use_streaming)

            join_dfs.append(df)

    return functools.reduce(
        lambda prev, curr: prev.join(curr, on=group_by, how="outer", coalesce=True),
        join_dfs,
    )


def card_df(
    set_code: str,
    extensions: list[ColumnSpec] | None = None,
):
    col_spec_map = dict(spells.columns.col_spec_map)
    if extensions is not None:
        for spec in extensions:
            col_spec_map[spec.name] = spec

    col_def_map = _hydrate_col_defs(set_code, col_spec_map)

    columns = [ColName.NAME] + [
        c for c, cdef in col_def_map.items() if cdef.col_type == ColType.CARD_ATTR
    ]
    fp = data_file_path(set_code, View.CARD)
    card_df = pl.read_parquet(fp)
    select_df = _view_select(
        card_df, frozenset(columns), col_def_map, is_agg_view=False
    )
    return select_df.select(columns)


def summon(
    set_code: str,
    columns: list[str] | None = None,
    group_by: list[str] | None = None,
    filter_spec: dict | None = None,
    extensions: list[ColumnSpec] | None = None,
    use_streaming: bool = False,
    read_cache: bool = True,
    write_cache: bool = True,
) -> pl.DataFrame:
    col_spec_map = dict(spells.columns.col_spec_map)
    if extensions is not None:
        for spec in extensions:
            col_spec_map[spec.name] = spec

    col_def_map = _hydrate_col_defs(set_code, col_spec_map)
    m = spells.manifest.create(col_def_map, columns, group_by, filter_spec)

    calc_fn = functools.partial(_base_agg_df, set_code, m, use_streaming=use_streaming)
    agg_df = _fetch_or_cache(
        calc_fn,
        set_code,
        (
            set_code,
            sorted(m.view_cols.get(View.DRAFT, set())),
            sorted(m.view_cols.get(View.GAME, set())),
            sorted(c.signature or "" for c in m.col_def_map.values()),
            sorted(m.base_view_group_by),
            filter_spec,
        ),
        read_cache=read_cache,
        write_cache=write_cache,
    )

    if View.CARD in m.view_cols:
        card_cols = m.view_cols[View.CARD].union({ColName.NAME})
        fp = data_file_path(set_code, View.CARD)
        card_df = pl.read_parquet(fp)
        select_df = _view_select(card_df, card_cols, m.col_def_map, is_agg_view=False)

        agg_df = agg_df.join(select_df, on="name", how="outer", coalesce=True)
        if ColName.NAME not in m.group_by:
            agg_df = agg_df.group_by(m.group_by).sum()

    ret_cols = m.group_by + m.columns
    ret_df = (
        _view_select(agg_df, frozenset(ret_cols), m.col_def_map, is_agg_view=True)
        .select(ret_cols)
        .sort(m.group_by)
    )

    return ret_df



from dataclasses import dataclass
from collections.abc import Callable

import polars as pl

from spells.enums import View, ColName, ColType


@dataclass(frozen=True)
class ColumnSpec:
    name: str
    col_type: ColType
    expr: pl.Expr | None = None
    exprMap: Callable[[str], pl.Expr] | None = None
    views: list[View] | None = None
    dependencies: list[str] | None = None
    version: str | None = (
        None  # only needed for user-defined functions with python functions in expr
    )


@dataclass(frozen=True)
class ColumnDefinition:
    name: str
    col_type: ColType
    expr: pl.Expr | tuple[pl.Expr, ...]
    views: tuple[View, ...]
    dependencies: tuple[str, ...]
    signature: str


default_columns = [
    ColName.COLOR,
    ColName.RARITY,
    ColName.NUM_SEEN,
    ColName.ALSA,
    ColName.NUM_TAKEN,
    ColName.ATA,
    ColName.NUM_GP,
    ColName.PCT_GP,
    ColName.GP_WR,
    ColName.NUM_OH,
    ColName.OH_WR,
    ColName.NUM_GIH,
    ColName.GIH_WR,
]

_column_specs = [
    ColumnSpec(
        name=ColName.NAME,
        col_type=ColType.GROUP_BY,
        # handled by internals, derived from both 'pick' and "name mapped" columns
    ),
    ColumnSpec(
        name=ColName.EXPANSION,
        col_type=ColType.GROUP_BY,
        views=[View.GAME, View.DRAFT],
    ),
    ColumnSpec(
        name=ColName.EVENT_TYPE,
        col_type=ColType.GROUP_BY,
        views=[View.GAME, View.DRAFT],
    ),
    ColumnSpec(
        name=ColName.DRAFT_ID,
        views=[View.GAME, View.DRAFT],
        col_type=ColType.FILTER_ONLY,
    ),
    ColumnSpec(
        name=ColName.DRAFT_TIME,
        col_type=ColType.FILTER_ONLY,
        views=[View.GAME, View.DRAFT],
    ),
    ColumnSpec(
        name=ColName.DRAFT_DATE,
        col_type=ColType.GROUP_BY,
        expr=pl.col("draft_time").str.to_datetime("%Y-%m-%d %H:%M:%S").dt.date(),
        dependencies=[ColName.DRAFT_TIME],
    ),
    ColumnSpec(
        name=ColName.DRAFT_DAY_OF_WEEK,
        col_type=ColType.GROUP_BY,
        expr=pl.col("draft_time").str.to_datetime("%Y-%m-%d %H:%M:%S").dt.weekday(),
        dependencies=[ColName.DRAFT_TIME],
    ),
    ColumnSpec(
        name=ColName.DRAFT_HOUR,
        col_type=ColType.GROUP_BY,
        expr=pl.col("draft_time").str.to_datetime("%Y-%m-%d %H:%M:%S").dt.hour(),
        dependencies=[ColName.DRAFT_TIME],
    ),
    ColumnSpec(
        name=ColName.DRAFT_WEEK,
        col_type=ColType.GROUP_BY,
        expr=pl.col("draft_time").str.to_datetime("%Y-%m-%d %H:%M:%S").dt.week(),
        dependencies=[ColName.DRAFT_TIME],
    ),
    ColumnSpec(
        name=ColName.RANK,
        col_type=ColType.GROUP_BY,
        views=[View.GAME, View.DRAFT],
    ),
    ColumnSpec(
        name=ColName.USER_N_GAMES_BUCKET,
        col_type=ColType.GROUP_BY,
        views=[View.DRAFT, View.GAME],
    ),
    ColumnSpec(
        name=ColName.USER_GAME_WIN_RATE_BUCKET,
        col_type=ColType.GROUP_BY,
        views=[View.DRAFT, View.GAME],
    ),
    ColumnSpec(
        name=ColName.PLAYER_COHORT,
        col_type=ColType.GROUP_BY,
        expr=pl.when(pl.col("user_n_games_bucket") < 100)
        .then(pl.lit("Other"))
        .otherwise(
            pl.when(pl.col("user_game_win_rate_bucket") > 0.57)
            .then(pl.lit("Top"))
            .otherwise(
                pl.when(pl.col("user_game_win_rate_bucket") < 0.49)
                .then(pl.lit("Bottom"))
                .otherwise(pl.lit("Middle"))
            )
        ),
        dependencies=[ColName.USER_N_GAMES_BUCKET, ColName.USER_GAME_WIN_RATE_BUCKET],
    ),
    ColumnSpec(
        name=ColName.EVENT_MATCH_WINS,
        col_type=ColType.GROUP_BY,
        views=[View.DRAFT],
    ),
    ColumnSpec(
        name=ColName.EVENT_MATCH_WINS_SUM,
        col_type=ColType.PICK_SUM,
        views=[View.DRAFT],
        expr=pl.col(ColName.EVENT_MATCH_WINS),
        dependencies=[ColName.EVENT_MATCH_WINS],
    ),
    ColumnSpec(
        name=ColName.EVENT_MATCH_LOSSES,
        col_type=ColType.GROUP_BY,
        views=[View.DRAFT],
    ),
    ColumnSpec(
        name=ColName.EVENT_MATCH_LOSSES_SUM,
        col_type=ColType.PICK_SUM,
        expr=pl.col(ColName.EVENT_MATCH_LOSSES),
        dependencies=[ColName.EVENT_MATCH_LOSSES],
    ),
    ColumnSpec(
        name=ColName.EVENT_MATCHES,
        col_type=ColType.GROUP_BY,
        expr=pl.col("event_match_wins") + pl.col("event_match_losses"),
        dependencies=[ColName.EVENT_MATCH_WINS, ColName.EVENT_MATCH_LOSSES],
    ),
    ColumnSpec(
        name=ColName.EVENT_MATCHES_SUM,
        col_type=ColType.PICK_SUM,
        expr=pl.col(ColName.EVENT_MATCHES),
        dependencies=[ColName.EVENT_MATCHES],
    ),
    ColumnSpec(
        name=ColName.IS_TROPHY,
        col_type=ColType.GROUP_BY,
        expr=pl.when(pl.col("event_type") == "Traditional")
        .then(pl.col("event_match_wins") == 3)
        .otherwise(pl.col("event_match_wins") == 7),
        dependencies=[ColName.EVENT_TYPE, ColName.EVENT_MATCH_WINS],
    ),
    ColumnSpec(
        name=ColName.IS_TROPHY_SUM,
        col_type=ColType.PICK_SUM,
        expr=pl.col(ColName.IS_TROPHY),
        dependencies=[ColName.IS_TROPHY],
    ),
    ColumnSpec(
        name=ColName.PACK_NUMBER,
        col_type=ColType.FILTER_ONLY,  # use pack_num
        views=[View.DRAFT],
    ),
    ColumnSpec(
        name=ColName.PACK_NUM,
        col_type=ColType.GROUP_BY,
        expr=pl.col("pack_number") + 1,
        dependencies=[ColName.PACK_NUMBER],
    ),
    ColumnSpec(
        name=ColName.PICK_NUMBER,
        col_type=ColType.FILTER_ONLY,  # use pick_num
        views=[View.DRAFT],
    ),
    ColumnSpec(
        name=ColName.PICK_NUM,
        col_type=ColType.GROUP_BY,
        expr=pl.col("pick_number") + 1,
        dependencies=[ColName.PICK_NUMBER],
    ),
    ColumnSpec(
        name=ColName.TAKEN_AT,
        col_type=ColType.PICK_SUM,
        expr=pl.col(ColName.PICK_NUM),
        dependencies=[ColName.PICK_NUM],
    ),
    ColumnSpec(
        name=ColName.NUM_TAKEN,
        col_type=ColType.PICK_SUM,
        expr=pl.when(pl.col(ColName.PICK).is_not_null())
        .then(1)
        .otherwise(0),  # a literal returns one row under select alone
        dependencies=[ColName.PICK],
    ),
    ColumnSpec(
        name=ColName.PICK,
        col_type=ColType.FILTER_ONLY,  # aggregated as "name"
        views=[View.DRAFT],
    ),
    ColumnSpec(
        name=ColName.PICK_MAINDECK_RATE,
        col_type=ColType.PICK_SUM,
        views=[View.DRAFT],
    ),
    ColumnSpec(
        name=ColName.PICK_SIDEBOARD_IN_RATE,
        col_type=ColType.PICK_SUM,
        views=[View.DRAFT],
    ),
    ColumnSpec(
        name=ColName.PACK_CARD,
        col_type=ColType.NAME_SUM,
        views=[View.DRAFT],
    ),
    ColumnSpec(
        name=ColName.LAST_SEEN,
        col_type=ColType.NAME_SUM,
        exprMap=lambda name: pl.col(f"pack_card_{name}")
        * pl.min_horizontal("pick_num", 8),
        dependencies=[ColName.PACK_CARD, ColName.PICK_NUM],
    ),
    ColumnSpec(
        name=ColName.NUM_SEEN,
        col_type=ColType.NAME_SUM,
        exprMap=lambda name: pl.col(f"pack_card_{name}") * (pl.col("pick_num") <= 8),
        dependencies=[ColName.PACK_CARD, ColName.PICK_NUM],
    ),
    ColumnSpec(
        name=ColName.POOL,
        col_type=ColType.NAME_SUM,
        views=[View.DRAFT],
    ),
    ColumnSpec(
        name=ColName.GAME_TIME,
        col_type=ColType.FILTER_ONLY,
        views=[View.GAME],
    ),
    ColumnSpec(
        name=ColName.GAME_DATE,
        col_type=ColType.GROUP_BY,
        expr=pl.col("game_time").str.to_datetime("%Y-%m-%d %H-%M-%S").dt.date(),
        dependencies=[ColName.GAME_TIME],
    ),
    ColumnSpec(
        name=ColName.GAME_DAY_OF_WEEK,
        col_type=ColType.GROUP_BY,
        expr=pl.col("game_time").str.to_datetime("%Y-%m-%d %H-%M-%S").dt.weekday(),
        dependencies=[ColName.GAME_TIME],
    ),
    ColumnSpec(
        name=ColName.GAME_HOUR,
        col_type=ColType.GROUP_BY,
        expr=pl.col("game_time").str.to_datetime("%Y-%m-%d %H-%M-%S").dt.hour(),
        dependencies=[ColName.GAME_TIME],
    ),
    ColumnSpec(
        name=ColName.GAME_WEEK,
        col_type=ColType.GROUP_BY,
        expr=pl.col("game_time").str.to_datetime("%Y-%m-%d %H-%M-%S").dt.week(),
        dependencies=[ColName.GAME_TIME],
    ),
    ColumnSpec(
        name=ColName.BUILD_INDEX,
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColumnSpec(
        name=ColName.MATCH_NUMBER,
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColumnSpec(
        name=ColName.GAME_NUMBER,
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColumnSpec(
        name=ColName.NUM_GAMES,
        col_type=ColType.GAME_SUM,
        expr=pl.col(ColName.GAME_NUMBER).is_not_null(),
        dependencies=[ColName.GAME_NUMBER],
    ),
    ColumnSpec(
        name=ColName.NUM_MATCHES,
        col_type=ColType.GAME_SUM,
        expr=pl.col(ColName.GAME_NUMBER) == 1,
        dependencies=[ColName.GAME_NUMBER],
    ),
    ColumnSpec(
        name=ColName.NUM_EVENTS,
        col_type=ColType.GAME_SUM,
        expr=(pl.col(ColName.GAME_NUMBER) == 1) & (pl.col(ColName.MATCH_NUMBER) == 1),
        dependencies=[ColName.GAME_NUMBER, ColName.MATCH_NUMBER],
    ),
    ColumnSpec(
        name=ColName.OPP_RANK,
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColumnSpec(
        name=ColName.MAIN_COLORS,
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColumnSpec(
        name=ColName.NUM_COLORS,
        col_type=ColType.GROUP_BY,
        expr=pl.col(ColName.MAIN_COLORS).str.len_chars(),
        dependencies=[ColName.MAIN_COLORS],
    ),
    ColumnSpec(
        name=ColName.SPLASH_COLORS,
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColumnSpec(
        name=ColName.HAS_SPLASH,
        col_type=ColType.GROUP_BY,
        expr=pl.col(ColName.SPLASH_COLORS).str.len_chars() > 0,
        dependencies=[ColName.SPLASH_COLORS],
    ),
    ColumnSpec(
        name=ColName.ON_PLAY,
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColumnSpec(
        name=ColName.NUM_ON_PLAY,
        col_type=ColType.GAME_SUM,
        expr=pl.col(ColName.ON_PLAY),
        dependencies=[ColName.ON_PLAY],
    ),
    ColumnSpec(
        name=ColName.NUM_MULLIGANS,
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColumnSpec(
        name=ColName.NUM_MULLIGANS_SUM,
        col_type=ColType.GAME_SUM,
        expr=pl.col(ColName.NUM_MULLIGANS),
        dependencies=[ColName.NUM_MULLIGANS],
    ),
    ColumnSpec(
        name=ColName.OPP_NUM_MULLIGANS,
        col_type=ColType.GAME_SUM,
        views=[View.GAME],
    ),
    ColumnSpec(
        name=ColName.OPP_NUM_MULLIGANS_SUM,
        col_type=ColType.GAME_SUM,
        expr=pl.col(ColName.OPP_NUM_MULLIGANS),
        dependencies=[ColName.OPP_NUM_MULLIGANS],
    ),
    ColumnSpec(
        name=ColName.OPP_COLORS,
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColumnSpec(
        name=ColName.NUM_TURNS,
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColumnSpec(
        name=ColName.NUM_TURNS_SUM,
        col_type=ColType.GAME_SUM,
        expr=pl.col(ColName.NUM_TURNS),
        dependencies=[ColName.NUM_TURNS],
    ),
    ColumnSpec(
        name=ColName.WON,
        col_type=ColType.GROUP_BY,
        views=[View.GAME],
    ),
    ColumnSpec(
        name=ColName.NUM_WON,
        col_type=ColType.GAME_SUM,
        expr=pl.col(ColName.WON),
        dependencies=[ColName.WON],
    ),
    ColumnSpec(
        name=ColName.OPENING_HAND,
        col_type=ColType.NAME_SUM,
        views=[View.GAME],
    ),
    ColumnSpec(
        name=ColName.WON_OPENING_HAND,
        col_type=ColType.NAME_SUM,
        exprMap=lambda name: pl.col(f"opening_hand_{name}") * pl.col(ColName.WON),
        dependencies=[ColName.OPENING_HAND, ColName.WON],
    ),
    ColumnSpec(
        name=ColName.DRAWN,
        col_type=ColType.NAME_SUM,
        views=[View.GAME],
    ),
    ColumnSpec(
        name=ColName.WON_DRAWN,
        col_type=ColType.NAME_SUM,
        exprMap=lambda name: pl.col(f"drawn_{name}") * pl.col(ColName.WON),
        dependencies=[ColName.DRAWN, ColName.WON],
    ),
    ColumnSpec(
        name=ColName.TUTORED,
        col_type=ColType.NAME_SUM,
        views=[View.GAME],
    ),
    ColumnSpec(
        name=ColName.WON_TUTORED,
        col_type=ColType.NAME_SUM,
        exprMap=lambda name: pl.col(f"tutored_{name}") * pl.col(ColName.WON),
        dependencies=[ColName.TUTORED, ColName.WON],
    ),
    ColumnSpec(
        name=ColName.DECK,
        col_type=ColType.NAME_SUM,
        views=[View.GAME],
    ),
    ColumnSpec(
        name=ColName.WON_DECK,
        col_type=ColType.NAME_SUM,
        exprMap=lambda name: pl.col(f"deck_{name}") * pl.col(ColName.WON),
        dependencies=[ColName.DECK, ColName.WON],
    ),
    ColumnSpec(
        name=ColName.SIDEBOARD,
        col_type=ColType.NAME_SUM,
        views=[View.GAME],
    ),
    ColumnSpec(
        name=ColName.WON_SIDEBOARD,
        col_type=ColType.NAME_SUM,
        exprMap=lambda name: pl.col(f"sideboard_{name}") * pl.col(ColName.WON),
        dependencies=[ColName.SIDEBOARD, ColName.WON],
    ),
    ColumnSpec(
        name=ColName.NUM_GNS,
        col_type=ColType.NAME_SUM,
        exprMap=lambda name: pl.max_horizontal(
            0,
            pl.col(f"deck_{name}")
            - pl.col(f"drawn_{name}")
            - pl.col(f"tutored_{name}")
            - pl.col(f"opening_hand_{name}"),
        ),
        dependencies=[
            ColName.DECK,
            ColName.DRAWN,
            ColName.TUTORED,
            ColName.OPENING_HAND,
        ],
    ),
    ColumnSpec(
        name=ColName.WON_NUM_GNS,
        col_type=ColType.NAME_SUM,
        exprMap=lambda name: pl.col(ColName.WON) * pl.col(f"num_gns_{name}"),
        dependencies=[ColName.NUM_GNS, ColName.WON],
    ),
    ColumnSpec(
        name=ColName.SET_CODE,
        col_type=ColType.CARD_ATTR,
    ),
    ColumnSpec(
        name=ColName.COLOR,
        col_type=ColType.CARD_ATTR,
    ),
    ColumnSpec(
        name=ColName.RARITY,
        col_type=ColType.CARD_ATTR,
    ),
    ColumnSpec(
        name=ColName.COLOR_IDENTITY,
        col_type=ColType.CARD_ATTR,
    ),
    ColumnSpec(
        name=ColName.CARD_TYPE,
        col_type=ColType.CARD_ATTR,
    ),
    ColumnSpec(
        name=ColName.SUBTYPE,
        col_type=ColType.CARD_ATTR,
    ),
    ColumnSpec(
        name=ColName.MANA_VALUE,
        col_type=ColType.CARD_ATTR,
    ),
    ColumnSpec(
        name=ColName.MANA_COST,
        col_type=ColType.CARD_ATTR,
    ),
    ColumnSpec(
        name=ColName.POWER,
        col_type=ColType.CARD_ATTR,
    ),
    ColumnSpec(
        name=ColName.TOUGHNESS,
        col_type=ColType.CARD_ATTR,
    ),
    ColumnSpec(
        name=ColName.IS_BONUS_SHEET,
        col_type=ColType.CARD_ATTR,
    ),
    ColumnSpec(
        name=ColName.IS_DFC,
        col_type=ColType.CARD_ATTR,
    ),
    ColumnSpec(
        name=ColName.ORACLE_TEXT,
        col_type=ColType.CARD_ATTR,
    ),
    ColumnSpec(
        name=ColName.CARD_JSON,
        col_type=ColType.CARD_ATTR,
    ),
    ColumnSpec(
        name=ColName.PICKED_MATCH_WR,
        col_type=ColType.AGG,
        expr=pl.col(ColName.EVENT_MATCH_WINS_SUM) / pl.col(ColName.EVENT_MATCHES),
        dependencies=[ColName.EVENT_MATCH_WINS_SUM, ColName.EVENT_MATCHES],
    ),
    ColumnSpec(
        name=ColName.TROPHY_RATE,
        col_type=ColType.AGG,
        expr=pl.col(ColName.IS_TROPHY_SUM) / pl.col(ColName.NUM_TAKEN),
        dependencies=[ColName.IS_TROPHY_SUM, ColName.NUM_TAKEN],
    ),
    ColumnSpec(
        name=ColName.GAME_WR,
        col_type=ColType.AGG,
        expr=pl.col(ColName.NUM_WON) / pl.col(ColName.NUM_GAMES),
        dependencies=[ColName.NUM_WON, ColName.NUM_GAMES],
    ),
    ColumnSpec(
        name=ColName.ALSA,
        col_type=ColType.AGG,
        expr=pl.col(ColName.LAST_SEEN) / pl.col(ColName.NUM_SEEN),
        dependencies=[ColName.LAST_SEEN, ColName.NUM_SEEN],
    ),
    ColumnSpec(
        name=ColName.ATA,
        col_type=ColType.AGG,
        expr=pl.col(ColName.TAKEN_AT) / pl.col(ColName.NUM_TAKEN),
        dependencies=[ColName.TAKEN_AT, ColName.NUM_TAKEN],
    ),
    ColumnSpec(
        name=ColName.NUM_GP,
        col_type=ColType.AGG,
        expr=pl.col(ColName.DECK),
        dependencies=[ColName.DECK],
    ),
    ColumnSpec(
        name=ColName.PCT_GP,
        col_type=ColType.AGG,
        expr=pl.col(ColName.DECK) / (pl.col(ColName.DECK) + pl.col(ColName.SIDEBOARD)),
        dependencies=[ColName.DECK, ColName.SIDEBOARD],
    ),
    ColumnSpec(
        name=ColName.GP_WR,
        col_type=ColType.AGG,
        expr=pl.col(ColName.WON_DECK) / pl.col(ColName.DECK),
        dependencies=[ColName.WON_DECK, ColName.DECK],
    ),
    ColumnSpec(
        name=ColName.NUM_OH,
        col_type=ColType.AGG,
        expr=pl.col(ColName.OPENING_HAND),
        dependencies=[ColName.OPENING_HAND],
    ),
    ColumnSpec(
        name=ColName.OH_WR,
        col_type=ColType.AGG,
        expr=pl.col(ColName.WON_OPENING_HAND) / pl.col(ColName.OPENING_HAND),
        dependencies=[ColName.WON_OPENING_HAND, ColName.OPENING_HAND],
    ),
    ColumnSpec(
        name=ColName.NUM_GIH,
        col_type=ColType.AGG,
        expr=pl.col(ColName.OPENING_HAND) + pl.col(ColName.DRAWN),
        dependencies=[ColName.OPENING_HAND, ColName.DRAWN],
    ),
    ColumnSpec(
        name=ColName.NUM_GIH_WON,
        col_type=ColType.AGG,
        expr=pl.col(ColName.WON_OPENING_HAND) + pl.col(ColName.WON_DRAWN),
        dependencies=[ColName.WON_OPENING_HAND, ColName.WON_DRAWN],
    ),
    ColumnSpec(
        name=ColName.GIH_WR,
        col_type=ColType.AGG,
        expr=pl.col(ColName.NUM_GIH_WON) / pl.col(ColName.NUM_GIH),
        dependencies=[ColName.NUM_GIH_WON, ColName.NUM_GIH],
    ),
    ColumnSpec(
        name=ColName.GNS_WR,
        col_type=ColType.AGG,
        expr=pl.col(ColName.WON_NUM_GNS) / pl.col(ColName.NUM_GNS),
        dependencies=[ColName.WON_NUM_GNS, ColName.NUM_GNS],
    ),
    ColumnSpec(
        name=ColName.IWD,
        col_type=ColType.AGG,
        expr=pl.col(ColName.GIH_WR) - pl.col(ColName.GNS_WR),
        dependencies=[ColName.GIH_WR, ColName.GNS_WR],
    ),
    ColumnSpec(
        name=ColName.NUM_IN_POOL,
        col_type=ColType.AGG,
        expr=pl.col(ColName.DECK) + pl.col(ColName.SIDEBOARD),
        dependencies=[ColName.DECK, ColName.SIDEBOARD],
    ),
    ColumnSpec(
        name=ColName.IN_POOL_WR,
        col_type=ColType.AGG,
        expr=(pl.col(ColName.WON_DECK) + pl.col(ColName.WON_SIDEBOARD))
        / pl.col(ColName.NUM_IN_POOL),
        dependencies=[ColName.WON_DECK, ColName.WON_SIDEBOARD, ColName.NUM_IN_POOL],
    ),
    ColumnSpec(
        name=ColName.DECK_TOTAL,
        col_type=ColType.AGG,
        expr=pl.col(ColName.DECK).sum(),
        dependencies=[ColName.DECK],
    ),
    ColumnSpec(
        name=ColName.WON_DECK_TOTAL,
        col_type=ColType.AGG,
        expr=pl.col(ColName.WON_DECK).sum(),
        dependencies=[ColName.WON_DECK],
    ),
    ColumnSpec(
        name=ColName.GP_WR_MEAN,
        col_type=ColType.AGG,
        expr=pl.col(ColName.WON_DECK_TOTAL) / pl.col(ColName.DECK_TOTAL),
        dependencies=[ColName.WON_DECK_TOTAL, ColName.DECK_TOTAL],
    ),
    ColumnSpec(
        name=ColName.GP_WR_EXCESS,
        col_type=ColType.AGG,
        expr=pl.col(ColName.GP_WR) - pl.col(ColName.GP_WR_MEAN),
        dependencies=[ColName.GP_WR, ColName.GP_WR_MEAN],
    ),
    ColumnSpec(
        name=ColName.GP_WR_VAR,
        col_type=ColType.AGG,
        expr=(pl.col(ColName.GP_WR_EXCESS).pow(2) * pl.col(ColName.NUM_GP)).sum()
        / pl.col(ColName.DECK_TOTAL),
        dependencies=[ColName.GP_WR_EXCESS, ColName.NUM_GP, ColName.DECK_TOTAL],
    ),
    ColumnSpec(
        name=ColName.GP_WR_STDEV,
        col_type=ColType.AGG,
        expr=pl.col(ColName.GP_WR_VAR).sqrt(),
        dependencies=[ColName.GP_WR_VAR],
    ),
    ColumnSpec(
        name=ColName.GP_WR_Z,
        col_type=ColType.AGG,
        expr=pl.col(ColName.GP_WR_EXCESS) / pl.col(ColName.GP_WR_STDEV),
        dependencies=[ColName.GP_WR_EXCESS, ColName.GP_WR_STDEV],
    ),
    ColumnSpec(
        name=ColName.GIH_TOTAL,
        col_type=ColType.AGG,
        expr=pl.col(ColName.NUM_GIH).sum(),
        dependencies=[ColName.NUM_GIH],
    ),
    ColumnSpec(
        name=ColName.WON_GIH_TOTAL,
        col_type=ColType.AGG,
        expr=pl.col(ColName.NUM_GIH_WON).sum(),
        dependencies=[ColName.NUM_GIH_WON],
    ),
    ColumnSpec(
        name=ColName.GIH_WR_MEAN,
        col_type=ColType.AGG,
        expr=pl.col(ColName.WON_GIH_TOTAL) / pl.col(ColName.GIH_TOTAL),
        dependencies=[ColName.WON_GIH_TOTAL, ColName.GIH_TOTAL],
    ),
    ColumnSpec(
        name=ColName.GIH_WR_EXCESS,
        col_type=ColType.AGG,
        expr=pl.col(ColName.GIH_WR) - pl.col(ColName.GIH_WR_MEAN),
        dependencies=[ColName.GIH_WR, ColName.GIH_WR_MEAN],
    ),
    ColumnSpec(
        name=ColName.GIH_WR_VAR,
        col_type=ColType.AGG,
        expr=(pl.col(ColName.GIH_WR_EXCESS).pow(2) * pl.col(ColName.NUM_GIH)).sum()
        / pl.col(ColName.GIH_TOTAL),
        dependencies=[ColName.GIH_WR_EXCESS, ColName.NUM_GIH, ColName.GIH_TOTAL],
    ),
    ColumnSpec(
        name=ColName.GIH_WR_STDEV,
        col_type=ColType.AGG,
        expr=pl.col(ColName.GIH_WR_VAR).sqrt(),
        dependencies=[ColName.GIH_WR_VAR],
    ),
    ColumnSpec(
        name=ColName.GIH_WR_Z,
        col_type=ColType.AGG,
        expr=pl.col(ColName.GIH_WR_EXCESS) / pl.col(ColName.GIH_WR_STDEV),
        dependencies=[ColName.GIH_WR_EXCESS, ColName.GIH_WR_STDEV],
    ),
]

col_spec_map = {col.name: col for col in _column_specs}

for item in ColName:
    assert item in col_spec_map, f"column {item} enumerated but not specified"
